#include <iostream>
#include <iomanip>
#include "Timer.h"
#include "Image.h"
#include "Color.h"
#include "Ray.h"
#include "Sphere.h"
#include "Point.h"
#include <glew.h>
#include <GLUT/glut.h>
#include <OPENGL/gl.h>
#include <OPENGL/glu.h>

using namespace std;

int writeSwitch = 1;		// switch to control writing image
int doOnce = 1;				// do something once
int mode = 1;				// switch for main rendering window

int xRes = 512;				// image X resolution
int yRes = 512;				// image Y resolution
Image myImage(xRes, yRes);	// the image to be written
int mainWindow;				// pointer to the main window

int currentStep = (yRes / 4);		// value to incriment with during progressive rendering

Sphere sphere1(Point(0.0, 0.2, 1.1), 1.0);		// the spheres to draw - global to avoid
Sphere sphere2(Point(1.4, 1.5, 1.2), 1.3);		// allocating structure more than once
Sphere sphere3(Point(-1.5, -0.5, 1.0), 1.9);	//

double t1, t2, t3, t4, r1, r2;		// values for time counters
double t5 = 0;						// counter for progressive rendering

char filename[] = "image1.ppm";

/**
*  Holds the raytracing functionality, nested loops that step
*  over each pixel in the image, or progressive using currentStep.
*/
void raytracer(void)
{
	r1 = Timer::currentSeconds();
	
	for(int i=0;i<yRes;i+=currentStep){ 
		for(int j=0;j<xRes;j++){ 
			Color result; 
			float x = 2 * (j - xRes/2. + 0.5) / xRes; 
			float y = 2 * (i - yRes/2. + 0.5) / yRes; 
			Ray ray(Point(0,0,-3), Vector(x, y, 1)); 
			if(sphere1.intersects(ray)) 
				result = Color(1, .4, 1); 
			else if(sphere2.intersects(ray)) 
				result = Color(.2, .3, 1); 
			else if(sphere3.intersects(ray)) 
				result = Color(1, .3, .2); 
			else 
				result = Color(.2, .1, .5); 
			myImage.set(j, i, result); 
		} 
	} 
	
	r2 = Timer::currentSeconds();
}

/**
*	Simple glut idle function.
*/
void myGlutIdle(void)
{
	glutPostRedisplay();
}

/**
*	Simple glut reshape function.
*
*	@param	x	the new x window coord
*	@param	y	the new y window coord
*/
void myGlutReshape(int	x, int y)
{
	glViewport(0, 0, xRes, yRes);
}

/**
*	GLUT keyboard control.  Pressing 'S' at any point
*	will overwrite and save the current render.
*
*	@param	key	the key pressed
*	@param	x	?
*	@param	y	?
*/
void myKeyboard(unsigned char key, int x, int y)
{
	switch(key)
	{
		case 's':	// save image
		case 'S':
			myImage.write(filename);
			break;
		case 27:	// ESC quits program
			exit(0);
			break;
		default:
			break;
	}
}

/**
*	Display function.  Draws the image pixels to the OpenGL window depending
*	on how much of the image is raytraced, or if currentStep is anything but 1,
*	which will progressively render the image.
*/
void myDisplay(void)
{
	glClear(GL_COLOR_BUFFER_BIT);

	raytracer();

	glDrawPixels(xRes, yRes, GL_RGB, GL_UNSIGNED_BYTE, myImage.imageData);
	
	glutSwapBuffers();
	glFlush();
	
	// logic to divide currentStep by half until it approaches 1
	currentStep = (currentStep <= 1) ? 1 : currentStep / 2;

	// prints some rendering statistics
	if(writeSwitch)
	{
		t5 += r2 - r1;
		cout << "Current Step:\t" << currentStep << "\n";
		cout << "Render time:\t" << t5 << "\n";
				
		if(currentStep == 1)
			writeSwitch = 0;	
	}
	else if(doOnce)
	{t5 += r2 - r1;
		cout << "\nTotal Render Time:\t" << t5 << "\n";
		doOnce = 0;
	}
}

/**
*	Main program, allows for two rendering options.  Option 1
*	renders the raytraces image completely, writes it to a file,
*	and then displays the result in an OpenGL window.  Option 2
*	raytraces and displays the image at the same time, rendering in
*	a progressive fashion.  Also contains the glutInit* and glut*Func
*	calls.
*/
int main (int argc, char** argv)
{
	cout << "Raytracer Terminal\n";
	cout << "1 : Render -> Write Image -> Display\n";
	cout << "2 : Render & Display Sequentially\n";
	cout << "S : Saves Current OpenGL Image\n\n";
	cin >> mode;
	
	t1 = Timer::currentSeconds();
	cout << "Rendering...\n";
	
	if (mode != 2)
	{
		t2 = Timer::currentSeconds();

		currentStep = 1;
		raytracer();
		cout << "Done\n--------------------\n";
		t3 = Timer::currentSeconds();
		myImage.write(filename);
		writeSwitch = 0;
		t4 = Timer::currentSeconds();
		
		cout << "Setup time:\t\t" << setprecision(3) << fixed << t2 - t1 << "\n";
		cout << "Render time:\t\t" << r2 - r1 << "\n";
		cout << "Image write time:\t" << t4 - t3 << "\n";
		cout << "Total Processing:\t" << t4 - t1 << "\n--------------------\n";
	}

	cout << "Displaying Image...\n";

	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);
	glutInitWindowSize(xRes, yRes);
	glutInitWindowPosition(100, 100);
	
	mainWindow = glutCreateWindow("Raytracer");
	glewInit();
	
	glClearColor(0.0, 0.0, 0.0, 0);
	
	glutIdleFunc(myGlutIdle);
	glutDisplayFunc(myDisplay);
	glutReshapeFunc(myGlutReshape);
	glutKeyboardFunc(myKeyboard);
	
	glutMainLoop();

    return 0;
}